package com.example.config.web;

import com.example.entity.common.RequestLogPo;
import com.example.service.common.RequestLogService;
import com.example.utils.AESUtil;
import com.example.utils.R;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.gson.Gson;
import com.google.gson.reflect.TypeToken;
import lombok.extern.slf4j.Slf4j;
import org.slf4j.MDC;
import org.springframework.lang.Nullable;
import org.springframework.util.StopWatch;
import org.springframework.util.StreamUtils;
import org.springframework.util.StringUtils;
import org.springframework.web.servlet.HandlerInterceptor;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.net.URLDecoder;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.UUID;

/**
 * 请求日志拦截器
 */
@Slf4j
public class RequestLogHandlerInterceptor implements HandlerInterceptor {
    /**
     * 计时器线程变量
     */
    private static final ThreadLocal<StopWatch> STOP_WATCH_THREAD_LOCAL = new ThreadLocal<>();
    private RequestLogService requestLogService;

    public RequestLogHandlerInterceptor(RequestLogService requestLogService) {
        this.requestLogService = requestLogService;
    }

    @Override
    public boolean preHandle(HttpServletRequest request, HttpServletResponse response, Object handler) throws Exception {
        StopWatch stopWatch = new StopWatch();
        stopWatch.start();
        // 计时器放入线程变量
        STOP_WATCH_THREAD_LOCAL.set(stopWatch);
        // 日志添加跟踪id
        MDC.put("traceId", UUID.randomUUID().toString().replace("-", ""));
        LocalDateTime now = LocalDateTime.now();
        log.info("请求地址: {} {}", request.getRequestURL().toString(), request.getMethod());
        // 请求头参数
        Map<String, String> headers = new HashMap<>();
        Enumeration<String> headerNames = request.getHeaderNames();
        while (headerNames.hasMoreElements()) {
            String k	= headerNames.nextElement();
            String v = request.getHeader(k);
            headers.put(k, v);
        }
        log.info("请求头参数: {}", new ObjectMapper().writeValueAsString(headers));
        // 查询参数
        Map<String, String> parameters = new HashMap<>();
        Enumeration<String> parameterNames = request.getParameterNames();
        while (parameterNames.hasMoreElements()) {
            String k	= parameterNames.nextElement();
            String v = request.getParameter(k);
            parameters.put(k, v);
        }
        log.info("查询参数: {}", new ObjectMapper().writeValueAsString(parameters));
        String body = null;
        if (!MyDispatcherServlet.UN_WRAPPER.contains(request.getRequestURI())) {
            // 请求体参数
            body = StreamUtils.copyToString(request.getInputStream(), Charset.forName(request.getCharacterEncoding()));
            log.info("请求体参数: {}", StringUtils.trimAllWhitespace(body));
        }
        String token = request.getHeader("token");
        if (!StringUtils.hasText(token) || "undefined".equals(token)) {
            token = parameters.get("token");
        }
        log.info("token: {}", token);
        if (!StringUtils.hasText(token)) {
            return forbidden(response);
        }
        try {
            String urlDecodeToken = URLDecoder.decode(token, StandardCharsets.UTF_8);
            log.info("urlDecodeToken: {}", urlDecodeToken);
            String decryptToken = AESUtil.decrypt(urlDecodeToken);
            log.info("decryptToken: {}", decryptToken);
            Gson gson = new Gson();
            Map<String, String> map = gson.fromJson(decryptToken, new TypeToken<Map<String, String>>() {
            }.getType());
            if (!"FINANCE".equals(map.get("APP_ID"))
                    || !StringUtils.hasText(map.get("LOGIN_ID"))
                    || !StringUtils.hasText(map.get("TIME_STAMP"))
                    || !StringUtils.hasText(map.get("EXPIRE_TIME"))
            ) {
                return forbidden(response);
            }
            LocalDateTime timeStamp = LocalDateTime.parse(map.get("TIME_STAMP"),
                    DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss"));
            int expireTime = Integer.parseInt(map.get("EXPIRE_TIME"));
            LocalDateTime expireTimeStamp = timeStamp.plusSeconds(expireTime);
            // token过期
            if (now.isAfter(expireTimeStamp)) {
                return unauthorized(response);
            }
            Map<String, String> requestParameters = new HashMap<>();
            requestParameters.put("request_parameters", gson.toJson(parameters));
            requestParameters.put("request_body", body);
            RequestLogPo requestLogPo = new RequestLogPo();
            requestLogPo.setRequestTime(now);
            requestLogPo.setLoginId(map.get("LOGIN_ID"));
            requestLogPo.setPageUrl(map.get("REQUEST_URL"));
            requestLogPo.setApi(request.getRequestURI());
            requestLogPo.setRequestParameters(gson.toJson(requestParameters));
            requestLogPo.setHeaders(gson.toJson(headers));
            requestLogPo.setAppId(map.get("APP_ID"));
            requestLogPo.setToken(urlDecodeToken);
            requestLogPo.setTimeStamp(timeStamp);
            requestLogPo.setExpireTime(expireTime);
            requestLogPo.setExpireTimeStamp(expireTimeStamp);
            requestLogService.offer(requestLogPo);
        } catch (Exception e) {
            log.error("token解密失败: {} {}", token, e.getMessage(), e);
            return forbidden(response);
        }
        return true;
    }

    @Override
    public void afterCompletion(HttpServletRequest request, HttpServletResponse response, Object handler,
                                @Nullable Exception ex) throws Exception {
        if (!MyDispatcherServlet.UN_WRAPPER.contains(request.getRequestURI())) {
            MyHttpServletResponseWrapper wrapper = (MyHttpServletResponseWrapper) response;
            String responseString = new String(wrapper.toByteArray());
//            log.info("响应 {}: {}", wrapper.getStatus(), responseString);
            // 返回结果打印前500个字符
            log.info("返回 {}: {}", wrapper.getStatus(),
                    org.apache.commons.lang3.StringUtils.substring(responseString, 0, 500));
        } else {
            log.info("响应 {}", response.getStatus());
        }
        StopWatch stopWatch = STOP_WATCH_THREAD_LOCAL.get();
        stopWatch.stop();
        log.info("耗时 {} ms", stopWatch.getTotalTimeMillis());
        STOP_WATCH_THREAD_LOCAL.remove();
    }

    private boolean forbidden(HttpServletResponse response) throws IOException {
        response.setCharacterEncoding("UTF-8");
        response.setContentType("application/json; charset=utf-8");
        String r = new Gson().toJson(R.error(403, "没有权限访问"));
        byte[] bytes = r.getBytes();
        response.getOutputStream().write(bytes);
        StopWatch stopWatch = STOP_WATCH_THREAD_LOCAL.get();
        stopWatch.stop();
        log.warn("返回 {}: {}", response.getStatus(), r);
        log.info("耗时 {} ms", stopWatch.getTotalTimeMillis());
        STOP_WATCH_THREAD_LOCAL.remove();
        return false;
    }

    private boolean unauthorized(HttpServletResponse response) throws IOException {
        response.setCharacterEncoding("UTF-8");
        response.setContentType("application/json; charset=utf-8");
        String r = new Gson().toJson(R.error(401, "请刷新页面或重新登录"));
        byte[] bytes = r.getBytes();
        response.getOutputStream().write(bytes);
        StopWatch stopWatch = STOP_WATCH_THREAD_LOCAL.get();
        stopWatch.stop();
        log.info("返回 {}: {}", response.getStatus(), r);
        log.info("耗时 {} ms", stopWatch.getTotalTimeMillis());
        STOP_WATCH_THREAD_LOCAL.remove();
        return false;
    }
}